using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Threading;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using Microsoft.CodeAnalysis.Diagnostics;

namespace FeedLibraryAnalyzer
{
    [DiagnosticAnalyzer(LanguageNames.CSharp)]
    public class FeedLibraryAnalyzerAnalyzer : DiagnosticAnalyzer
    {
        public const string DiagnosticId = "RSS001";
        internal static readonly string Title = "RSS URL analysis";
        internal static readonly string MessageFormat = "URL is invalid";
        internal static readonly string Description = "Provides live analysis for the FeedLibrary APIs";
        internal const string Category = "Syntax";


        private static DiagnosticDescriptor Rule = new DiagnosticDescriptor(DiagnosticId, Title, MessageFormat, Category, DiagnosticSeverity.Warning, isEnabledByDefault: true, description: Description);

        public override ImmutableArray<DiagnosticDescriptor> SupportedDiagnostics { get { return ImmutableArray.Create(Rule); } }

        public override void Initialize(AnalysisContext context)
        {
            // Register an action when compilation starts
            context.
                RegisterCompilationStartAction((CompilationStartAnalysisContext ctx) =>
            {
                //Detect if the type metadata
                //exists in the compilation context
                var myLibraryType = 
                ctx.Compilation.
                GetTypeByMetadataName("FeedLibrary.FeedItem");

                //If not, return
                if (myLibraryType == null)
                    return;

                //Register an action against
                //an InvocationExpression
                ctx.RegisterSyntaxNodeAction(AnalyzeMethodInvocation, 
                    SyntaxKind.InvocationExpression);
            });
        }

        private static void AnalyzeMethodInvocation(SyntaxNodeAnalysisContext context)
        {
            //Convert the current syntax node into an InvocationExpressionSyntax
            //which represents a method call
            var invocationExpr = (InvocationExpressionSyntax)context.Node;

            //Convert the associated expression into a MemberAccessExpressionSyntax
            //which represents a method's information
            //If the expression is not a MemberAccessExpressionSyntax, return
            if (!(invocationExpr.Expression is MemberAccessExpressionSyntax))
            {
                return;
            }
            var memberAccessExpr = (MemberAccessExpressionSyntax)invocationExpr.Expression;

            //If the method name is not ParseFeedAsync, return
            if (memberAccessExpr?.Name.ToString() != "ParseFeedAsync") { return; }

            //If the method name is ParseFeedAsync, check for the symbol
            //info and see if the return type matches
            var memberSymbol = context.SemanticModel.
                               GetSymbolInfo(memberAccessExpr).
                               Symbol as IMethodSymbol;
            if (memberSymbol == null) { return; }
            var result = memberSymbol.ToString();

            if (memberSymbol?.ReturnType.ToString() !=
            "System.Threading.Tasks.Task<System.Collections.Generic.IEnumerable<FeedLibrary.FeedItem>>")
            {
                return;
            }

            //Check if the method call has the required argument number
            var argumentList = invocationExpr.ArgumentList;
            if (argumentList?.Arguments.Count != 1) {
                return; }

            //Convert the expression for the first method argument into
            //a LiteralExpressionSyntax. If null, return
            var urlLiteral = (LiteralExpressionSyntax)invocationExpr.ArgumentList.
                Arguments[0].Expression;
            if (urlLiteral == null) { return; }

            //Convert the actual value for the method argument into string
            //If null, return
            var urlLiteralOpt = context.SemanticModel.GetConstantValue(urlLiteral);

            var urlValue = (string)urlLiteralOpt.Value;
            if (urlValue == null) { return; }

            //If the URL is not well-formed, create a diagnostic
            if (Uri.IsWellFormedUriString(urlValue, UriKind.Absolute) == false)
            {
                var diagn = Diagnostic.Create(Rule, urlLiteral.GetLocation(),
                    "The specified parameter Is Not a valid RSS feed");
                context.ReportDiagnostic(diagn);
            }
        }

    }
}
